﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Random;

namespace HeuristicLab.Problems.GrammaticalEvolution {
  /// <summary>
  /// DepthFirstMapper
  /// </summary>
  [Item("DepthFirstMapper", "Resolves the non-terminal symbols of the resulting phenotypic syntax tree in a depth-first manner.")]
  [StorableClass]
  public class DepthFirstMapper : GenotypeToPhenotypeMapper {

    [StorableConstructor]
    protected DepthFirstMapper(bool deserializing) : base(deserializing) { }
    protected DepthFirstMapper(DepthFirstMapper original, Cloner cloner) : base(original, cloner) { }
    public DepthFirstMapper() : base() { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new DepthFirstMapper(this, cloner);
    }


    /// <summary>
    /// Maps a genotype (an integer vector) to a phenotype (a symbolic expression tree).
    /// Depth-first approach.
    /// </summary>
    /// <param name="grammar">grammar definition</param>
    /// <param name="genotype">integer vector, which should be mapped to a tree</param>
    /// <returns>phenotype (a symbolic expression tree)</returns>
    public override SymbolicExpressionTree Map(ISymbolicExpressionGrammar grammar,
                                               IntegerVector genotype) {

      SymbolicExpressionTree tree = new SymbolicExpressionTree();
      var rootNode = (SymbolicExpressionTreeTopLevelNode)grammar.ProgramRootSymbol.CreateTreeNode();
      if (rootNode.HasLocalParameters) rootNode.ResetLocalParameters(new MersenneTwister());
      var startNode = (SymbolicExpressionTreeTopLevelNode)grammar.StartSymbol.CreateTreeNode();
      if (startNode.HasLocalParameters) startNode.ResetLocalParameters(new MersenneTwister());
      rootNode.AddSubtree(startNode);
      tree.Root = rootNode;

      //int genotypeIndex = 0;
      //int currSubtreeCount = 1;
      //MapDepthFirstRecursively(startNode, genotype,
      //                         grammar, genotype.Length,
      //                         ref genotypeIndex, ref currSubtreeCount,
      //                         new MersenneTwister());

      MapDepthFirstIteratively(startNode, genotype, grammar,
                               genotype.Length, new MersenneTwister());
      return tree;
    }


    /// <summary>
    /// Genotype-to-Phenotype mapper (recursive depth-first approach).
    /// Appends maximum allowed children (non-terminal symbols) to
    /// <paramref name="currentNode"/>, as long as <paramref name="currSubtreeCount"/>
    /// doesn't exceed <paramref name="maxSubtreeCount"/>.
    /// If at most <paramref name="maxSubtreeCount"/> subtrees were created,
    /// each non-full node is filled with randomly chosen nodes
    /// (non-terminal and terminal), and each non-terminal node is again filled with a terminal node.
    /// </summary>
    /// <param name="currentNode">current parent node</param>
    /// <param name="genotype">integer vector, which should be mapped to a tree</param>
    /// <param name="grammar">grammar to determine the allowed child symbols for currentNode </param>
    /// <param name="maxSubtreeCount">maximum allowed subtrees (= number of used genomes)</param>
    /// <param name="genotypeIndex">current index in integer vector</param>
    /// <param name="currSubtreeCount">number of already determined subtrees (filled or still incomplete)</param>
    private void MapDepthFirstRecursively(ISymbolicExpressionTreeNode currentNode,
                                          IntegerVector genotype,
                                          ISymbolicExpressionGrammar grammar,
                                          int maxSubtreeCount,
                                          ref int genotypeIndex,
                                          ref int currSubtreeCount,
                                          IRandom random) {

      // TODO: check, if method calls of GetNewChildNode() and GetRandomTerminalNode() don't return null
      if (currSubtreeCount < maxSubtreeCount) {

        var newNode = GetNewChildNode(currentNode, genotype, grammar, genotypeIndex, random);

        if ((currSubtreeCount + newNode.Symbol.MinimumArity) > maxSubtreeCount) {
          // TODO: maybe check, if there is any node, which fits in the tree yet
          currentNode.AddSubtree(GetRandomTerminalNode(currentNode, grammar, random));
        } else {
          currentNode.AddSubtree(newNode);
          genotypeIndex++;
          currSubtreeCount += newNode.Symbol.MinimumArity;

          while (newNode.Symbol.MinimumArity > newNode.SubtreeCount) {
            MapDepthFirstRecursively(newNode, genotype,
                                     grammar, maxSubtreeCount,
                                     ref genotypeIndex, ref currSubtreeCount, random);
          }
        }

      } else {
        while (currentNode.Symbol.MinimumArity > currentNode.SubtreeCount) {
          var newNode = GetNewChildNode(currentNode, genotype, grammar, genotypeIndex, random);
          currentNode.AddSubtree(newNode);
          genotypeIndex++;
          while (newNode.Symbol.MinimumArity > newNode.SubtreeCount) {
            newNode.AddSubtree(GetRandomTerminalNode(newNode, grammar, random));
          }
        }
      }
    }


    /// <summary>
    /// Genotype-to-Phenotype mapper (iterative depth-first approach, by using a stack -> LIFO).
    /// </summary>
    /// <param name="startNode">first node of the tree with arity 1</param>
    /// <param name="genotype">integer vector, which should be mapped to a tree</param>
    /// <param name="grammar">grammar to determine the allowed child symbols for each node</param>
    /// <param name="maxSubtreeCount">maximum allowed subtrees (= number of used genomes)</param>
    /// <param name="random">random number generator</param>
    private void MapDepthFirstIteratively(ISymbolicExpressionTreeNode startNode,
                                          IntegerVector genotype,
                                          ISymbolicExpressionGrammar grammar,
                                          int maxSubtreeCount, IRandom random) {

      Stack<Tuple<ISymbolicExpressionTreeNode, int>> stack
        = new Stack<Tuple<ISymbolicExpressionTreeNode, int>>(); // tuples of <node, arity>

      int genotypeIndex = 0;
      int currSubtreeCount = 1;

      stack.Push(new Tuple<ISymbolicExpressionTreeNode, int>(startNode, 1));

      while ((currSubtreeCount < maxSubtreeCount) && (stack.Count > 0)) {

        // get next node from stack and re-push it, if this node still has unhandled subtrees ...
        Tuple<ISymbolicExpressionTreeNode, int> current = stack.Pop();
        if (current.Item2 > 1) {
          stack.Push(new Tuple<ISymbolicExpressionTreeNode, int>(current.Item1, current.Item2 - 1));
        }

        var newNode = GetNewChildNode(current.Item1, genotype, grammar, genotypeIndex, random);
        int arity = SampleArity(random, newNode, maxSubtreeCount - currSubtreeCount);

        if (arity < 0) {
          current.Item1.AddSubtree(GetRandomTerminalNode(current.Item1, grammar, random));
        } else {
          current.Item1.AddSubtree(newNode);
          genotypeIndex++;
          currSubtreeCount += arity;
          if (arity > 0) {
            // new node has subtrees so push it onto the stack 
            stack.Push(new Tuple<ISymbolicExpressionTreeNode, int>(newNode, arity));
          }
        }
      }

      // maximum allowed subtree count was already reached, but there are still
      // incomplete subtrees (non-terminal symbols) in the tree
      // -> fill them with terminal symbols
      while (stack.Count > 0) {
        Tuple<ISymbolicExpressionTreeNode, int> current = stack.Pop();
        if (current.Item2 > 1) {
          stack.Push(new Tuple<ISymbolicExpressionTreeNode, int>(current.Item1, current.Item2 - 1));
        }
        current.Item1.AddSubtree(GetRandomTerminalNode(current.Item1, grammar, random));
      }
    }
  }
}